// @ts-nocheck
import { getAbsoluteCoordinate, getInitialPoint } from './annotation/utils.js';
import annotationStore from './annotation_store';

export default class Separation {
  documentViewer: any
  pageViewer: any
  container: HTMLDivElement
  pagePtr: number

  iccData: ArrayBuffer | Uint8Array = null
  isClearData: boolean = false
  private latestRequestPoint: Point | null = null
  private isRequestInProgress: boolean = false
  colorList: Array = []

  onMousemove = this.handleMouseMove.bind(this)
  onClick = this.handleClick.bind(this)
  onUpdateCanvas = this.updateCanvas.bind(this)

  constructor(options: {
    documentViewer: any,
    container: HTMLDivElement,
    pageViewer: any,
    pagePtr: number
  }) {
    this.documentViewer = options.documentViewer
    this.container = options.container
    this.pageViewer = options.pageViewer
    this.pagePtr = options.pagePtr
  }

  async init() {
    await this.initColorsList()

    this.documentViewer.eventBus._on('updateSeparationView', this.onUpdateCanvas)
    // this.container.addEventListener('click', this.onClick)
    document.addEventListener('mousemove', this.onMousemove)
  }

  async initColorsList() {
    const { messageHandler, doc, eventBus, colorFile } = this.documentViewer

    messageHandler.sendWithPromise('InitColorList', doc)
    const sepColorPtr = await messageHandler.sendWithPromise('GetPDFSeparationColor', {
      pagePtr: this.pagePtr.pagePtr,
      fileData: colorFile
    })
    this.pagePtr.sepColorPtr = sepColorPtr
    
    this.colorList = await messageHandler.sendWithPromise('GetColorsList', doc)
    eventBus.dispatch('colorSeparationUpdate', this.colorList)

    if (!this.documentViewer.colorList.length && this.colorList.length) {
      this.documentViewer.colorList = this.colorList
      setTimeout(() => {
        eventBus.dispatch('colorSeparationAdded', this.colorList)
      }, 100)
    }
  }

  destroy() {
    this.documentViewer.eventBus._off('updateSeparationView', this.onUpdateCanvas)
    // this.container.removeEventListener('click', this.onClick)
    document.removeEventListener('mousemove', this.onMousemove)
  }

  async handleMouseMove(event: MouseEvent) {
    if (this.documentViewer.doc !== this.pagePtr.doc) return
    const inPage = this.testPoint(event)
    if (!inPage) {
      !this.isClearData && this.documentViewer.eventBus.dispatch('colorSeparationUpdate', this.colorList)
      this.isClearData = true
      return
    }

    const { x, y } = getAbsoluteCoordinate(this.container, event)
    const point = getInitialPoint({ x, y }, this.pageViewer.viewport.scale)

    this.latestRequestPoint = point
    if (this.isRequestInProgress) return
    await this.processLatestRequest()
  }

  async handleClick(event: MouseEvent) {
    const { x, y } = getAbsoluteCoordinate(this.container, event)
    const point = getInitialPoint({ x, y }, this.pageViewer.viewport.scale)
  }

  async getSelectPointColor(point: Point) {
    const pointColors = await this.documentViewer.messageHandler.sendWithPromise('GetSelectPointColor', {
      pagePtr: this.pagePtr.pagePtr,
      sepColorPtr: this.pagePtr.sepColorPtr,
      point
    })

    if (pointColors.length) {
      const pointColorList = this.colorList.map(item => ({ ...item }))
      const colorValueList = []

      pointColors.forEach(color => {
        pointColorList[color.Index].intensity = color.ColorIntensity
        
        colorValueList.push({
          name: pointColorList[color.Index].name,
          value: color.ColorIntensity
        })
      })

      this.documentViewer.eventBus.dispatch('colorSeparationUpdate', pointColorList)
      this.isClearData = false

      return colorValueList
    } else {
      if (!this.isClearData) {
        this.documentViewer.eventBus.dispatch('colorSeparationUpdate', this.colorList)
      }
      this.isClearData = true

      return pointColors
    }
  }

  async updateCanvas(indexs) {
    if (!this.pageViewer.canvas) return

    const ratio = window.devicePixelRatio || 1
    const { width, height } = this.pageViewer.viewport
    const canvasWidth = Math.round(width * ratio)
    const canvasHeight = Math.round(height * ratio)

    const { imageArray } = await this.documentViewer.messageHandler.sendWithPromise('RenderPageAndIgnoreColorByIndex', {
      sepColorPtr: this.pagePtr.sepColorPtr,
      width: canvasWidth,
      height: canvasHeight,
      indexs
    })

    const pageCtx = this.pageViewer.canvas?.getContext('2d')
    if (!pageCtx) return
    pageCtx.clearRect(0, 0, canvasWidth, canvasHeight)
    const imageData = pageCtx.createImageData(canvasWidth, canvasHeight)
    imageData.data.set(imageArray)
    pageCtx.putImageData(imageData, 0, 0)

    this.documentViewer.eventBus.dispatch('redrawMarkups', {
      pageIndex: this.pageViewer.pageIndex,
      isRestorePage: false
    })
    this.documentViewer.eventBus.dispatch('finishUpdateSeparationView')
  }

  testPoint(event: MouseEvent) {
    const { x, y, width, height } = getAbsoluteCoordinate(this.container, event)
    if (x < 0 || y < 0 || x > width || y > height) return false
    return {
      x,
      y
    }
  }

  private async processLatestRequest() {
    this.isRequestInProgress = true

    const point = this.latestRequestPoint

    if (point) {
      try {
        await this.getSelectPointColor(point)
      } catch (error) {
        console.error('请求出错:', error)
      }
    }

    this.isRequestInProgress = false
  }
}
