import {
  getAbsoluteCoordinate,
  createElement,
  getInitialPoint
} from '../utils';

export default class PaintRedaction {
  constructor ({
    tool,
    color,
    layer,
  }) {
    this.layer = layer
    this._tool = tool
    this.color = color
    this.path = []

    this.initStartPoint = null
    this.initEndPoint = null
    this.linkEle = null

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)

    this.init()
  }

  init () {
    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)

    this.layer.pageDiv.addEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.addEventListener('touchstart', this.onMousedown)

    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
  }

  get tool () {
    return this._tool
  }

  set tool (toolType) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
      this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
      this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)
    }
    this._tool = toolType
  }

  reset () {
    this.initStartPoint = null
    this.initEndPoint = null
    if (this.linkEle) this.linkEle.remove()
    this.linkEle = null
    this.layer.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)

    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)
  }

  handleMouseDown (event) {
    if (this.layer.annotationStore.selectedElementName || this.layer.pageDiv.classList.contains('text')) return
    const { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    this.initStartPoint = { x, y }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  }

  handleMouseUp () {
    let initStartPoint = this.initStartPoint
    let initEndPoint = this.initEndPoint
    if (initEndPoint && initStartPoint.x !== initEndPoint.x && initStartPoint.y !== initEndPoint.y) {
      let start = getInitialPoint(initStartPoint, this.layer.viewport.scale)
      let end = getInitialPoint(initEndPoint, this.layer.viewport.scale)

      const newStart = {
        x: Math.min(start.x, end.x),
        y: Math.min(start.y, end.y)
      }
      const newEnd = {
        x: Math.max(start.x, end.x),
        y: Math.max(start.y, end.y)
      }
      const rect = {
        left: newStart.x,
        top: newStart.y,
        right: newEnd.x,
        bottom: newEnd.y
      }

      const annotationData = {
        rect,
        operate: 'add-annot',
        type: 'redact',
        pageIndex: this.layer.page,
        date: new Date(),
        erasure: this._tool === 'remove'
      }
      this._tool === 'redaction' && (annotationData.fillColor = 'rgb(0, 0, 0)')

      this.layer.eventBus.dispatch('annotationChanged', {
        type: 'add',
        show: true,
        annotation: annotationData
      })
    }
    // this.reset()
    
    this.initStartPoint = null
    this.initEndPoint = null
    if (this.linkEle) this.linkEle.remove()
    this.linkEle = null
    this.layer.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
  }

  handleMouseMove (event) {
    let { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    const { width, height } = this.layer.viewport
    const scale = this.layer.viewport.scale
    x = Math.min(width, x)
    x = Math.max(0, x)
    y = Math.min(height, y)
    y = Math.max(0, y)
    this.initEndPoint = { x, y }
    if (this.initStartPoint.x !== this.initEndPoint.x && this.initStartPoint.y !== this.initEndPoint.y) {
      if (!this.linkEle) {
        const linkEle = createElement('div', {
          'display': 'block',
          'border': `${2 * scale}px solid #FF0000`,
          'position': 'absolute'
        })
        this.linkEle = linkEle
        this.scene = this.layer.pageDiv.querySelector('.annotationLayer')
        this.scene.append(this.linkEle)
      } else {
        // if (event.shiftKey) {
        //   const { initStartPoint, initEndPoint } = this
        //   const distanceX = Math.abs(initEndPoint.x - initStartPoint.x)
        //   const distanceY = Math.abs(initEndPoint.y - initStartPoint.y)
        //   const distance = Math.min(distanceX, distanceY)
        //   if (initStartPoint.x < initEndPoint.x) {
        //     initEndPoint.x = initStartPoint.x + distance
        //   } else {
        //     initEndPoint.x = initStartPoint.x - distance
        //   }
        //   if (initStartPoint.y < initEndPoint.y) {
        //     initEndPoint.y = initStartPoint.y + distance
        //   } else {
        //     initEndPoint.y = initStartPoint.y - distance
        //   }
        // }
        const rect = this.calculate(this.initStartPoint, this.initEndPoint)

        const linkEle = this.linkEle
        linkEle.style.top = `${rect.top}px`;
        linkEle.style.left = `${rect.left}px`;
        linkEle.style.width = `${Math.abs(rect.width - 1)}px`;
        linkEle.style.height = `${Math.abs(rect.height - 1)}px`;
      }
    }
  }
  
  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });

    return {
      top: initRect.top,
      left: initRect.left,
      width: initRect.width,
      height: initRect.height,
    }
  }
}
