import Base from './base';
import { getActualPoint, getClickPoint, createSvg, createElement } from './utils';
import { onClickOutsideUp } from '../ui_utils'

export default class Polygon extends Base {
  
  constructor ({
    annotation,
    layer,
    show = false,
    enableReply = true
  }) {
    super({
      annotation,
      layer,
    })
    this.enableReply = enableReply

    this.hidden = true
    this.initial = false

    this.start = null
    this.end = null

    this.newStart = null
    this.newEnd = null

    this.svgElement = null

    this.startCircle = null
    this.endCircle = null

    this.show = show
    this.ratio = window.devicePixelRatio || 1
    this.decimalPlaces = 0 // 保留小数位
    this.showPerimeter = true
    this.showArea = true
    this.isPopupShow = false

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onOpenReply = this.openReply.bind(this)
    this.handlePropertyPanel = this.handlePropertyPanel.bind(this)
    this.onHandlePopup = this.handlePopup.bind(this)
    this.onKeydown = this.handleKeyDown.bind(this)
    this.render()
  }

  async render () {
    const annotation = this.annotation

    if (annotation.showPerimeter !== undefined) {
      this.showPerimeter = annotation.showPerimeter
      delete annotation.showPerimeter
    }
    if (annotation.showArea !== undefined) {
      this.showArea = annotation.showArea
      delete annotation.showArea
    }
    if (!annotation.fillColor) {
      annotation.fillColor = ''
    }
    if (!annotation.borderStyle) {
      annotation.borderStyle = 'solid'
    }
    if (!annotation.contents) {
      await this.getMeasureData()
      const textList = []
      this.showPerimeter && textList.push(this.annotation.perimeter + this.annotation.scale.translateUnit)
      this.showArea && textList.push(this.annotation.area + this.annotation.scale.translateUnit + '\u00B2')
      annotation.contents = textList.join('\n')
    }
    if (annotation.precision) {
      if (annotation.precision > 1) {
        annotation.precision = Math.pow(10, -Math.floor(Math.log10(annotation.precision)))
      }
      this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
      annotation.area = this.formatValue(annotation.area)
      annotation.perimeter = this.formatValue(annotation.perimeter)
    }
    if (!annotation.fontStyle) {
      annotation.fontStyle = 'regular'
    }

    const scale = this.layer.viewport.scale
    const { start, end } = this.getActualRect(scale)
    this.start = start
    this.end = end

    const rect = this.calculate(start, end)
    this.rect = rect
    const actualbdwidth = (annotation.borderWidth || 2) * scale

    if (!annotation.borderColor) {
      annotation.borderColor = annotation.color || '#FF0000'
    }
    
    const annotationContainer = document.createElement('div')
    annotationContainer.id = annotation.name
    annotationContainer.className = 'annotation'
    annotationContainer.style.top = rect.top + 'px'
    annotationContainer.style.left = rect.left + 'px'
    annotationContainer.style.width = rect.width + 'px'
    annotationContainer.style.height = rect.height + 'px'
    this.annotationContainer = annotationContainer

    this.svgElement = document.createElementNS("http://www.w3.org/2000/svg", "svg");
    this.svgElement.setAttribute("viewBox", `0 0 ${rect.width} ${rect.height}`)
    this.svgElement.style.pointerEvents = 'all'
    this.svgElement.style.opacity = annotation.opacity

    let dashString = ''
    const scaleValue = Math.ceil(this.ratio * this.layer.viewport.scale)
    if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
      dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
    } else if (annotation.dashes.length > 2) {
      dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
    }

    this.shapeElement = createSvg(
      "rect",
      {
        x: actualbdwidth / 2,
        y: actualbdwidth / 2,
        width: `${Math.abs(rect.width - actualbdwidth)}px`,
        height: `${Math.abs(rect.height - actualbdwidth)}px`,
        stroke: annotation.borderColor,
        'stroke-width': actualbdwidth,
        'stroke-dasharray': dashString,
        // 'stroke-opacity': annotation.opacity || annotation.fillTransparency || 1,
        fill: annotation.fillColor ? annotation.fillColor : 'none',
        'fill-opacity': annotation.fillTransparency || 1,
      }, {
        'stroke-linecap': 'butt'
      }
    )
    this.svgElement.append(this.shapeElement)
    this.annotationContainer.append(this.svgElement)
    this.layer.div.append(this.annotationContainer)
    this.svgElement.addEventListener('mousedown', this.handleClick.bind(this))

    if (annotation.measure) {
      this.textEle = createSvg('text', {
        'fill': annotation.color || annotation.borderColor,
        'text-anchor': 'middle',
        'dominant-baseline': 'middle',
        'font-size': annotation.fontSize + 'px'
      })
      this.svgElement.append(this.textEle)

      if (this.showPerimeter) {
        this.permiterTspan = createSvg('tspan', {
          x: (end.x - start.x) / 2,
          y: (end.y - start.y) / 2,
          dy: '0em'
        })
        this.permiterTspan.textContent = annotation.perimeter + annotation.scale.translateUnit
        this.textEle.appendChild(this.permiterTspan)
      }
      if (this.showArea) {
        this.areaTspan = createSvg('tspan', {
          x: (end.x - start.x) / 2,
          y: (end.y - start.y) / 2,
          dy: '1em'
        })
        this.areaTspan.textContent = annotation.area + annotation.scale.translateUnit + '\u00B2'
        this.textEle.appendChild(this.areaTspan)
      }
    }

    this.outerLineContainer = document.createElement('div')
    this.outerLineContainer.className = 'outline-container'
    
    const outerLine = document.createElementNS("http://www.w3.org/2000/svg", "svg");

    outerLine.style.position = 'absolute'
    outerLine.style.zIndex = 1
    outerLine.style.left = `${rect.left - 7}px`
    outerLine.style.top = `${rect.top - 7}px`
    outerLine.style.width = `${rect.width + 7 * 2}px`
    outerLine.style.height = `${rect.height + 7 * 2}px`
    outerLine.style.pointerEvents = 'all'
    this.outerLine = outerLine
    
    this.moveRect = createSvg(
      "rect",
      {
        class: "move",
        'data-id': "move",
        stroke: "#1460F3",
        'stroke-width': 1,
        'stroke-dasharray': '2',
        'fill-opacity': 0,
        width: rect.width + 8,
        height: rect.height + 8,
        x: 3,
        y: 3
      }
    );
    
    this.topLeftRect = createSvg(
      "rect",
      {
        class: "nw-resize",
        'data-id': "nw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: 0,
        y: 0
      }
    );

    this.bottomLeftRect = createSvg(
      "rect",
      {
        class: "sw-resize",
        'data-id': "sw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: 0,
        y: rect.height + 8
      }
    );
    
    this.bottomRightRect = createSvg(
      "rect",
      {
        class: "se-resize",
        'data-id': "se-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: rect.height + 8
      }
    );

    this.topRightRect = createSvg(
      "rect",
      {
        class: "ne-resize",
        'data-id': "ne-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: 0
      }
    );

    this.outerLine.append(this.moveRect)
    this.outerLine.append(this.topLeftRect)
    this.outerLine.append(this.bottomLeftRect)
    this.outerLine.append(this.bottomRightRect)
    this.outerLine.append(this.topRightRect)
    this.outerLineContainer.append(this.outerLine)
    this.initial = true
    if (this.show) {
      !this.layer.addMode && this.selectAnnotation()
      this.show = false
    }
  }

  getActualRect (s) {
    const annotation = this.annotation
    const { left: x1, top: y1, right: x2, bottom: y2 } = annotation.rect
    
    const start = getActualPoint(
      {
        x: x1,
        y: y1
      },
      s
    )
    const end = getActualPoint(
      {
        x: x2,
        y: y2
      },
      s
    )
    return {
      start,
      end
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = (this.annotation.borderWidth || 2) * this.layer.viewport.scale

    return {
      top: initRect.top - actualbdwidth,
      left: initRect.left - actualbdwidth,
      width: initRect.width + actualbdwidth * 2,
      height: initRect.height + actualbdwidth * 2,
    }
  }

  getInitialPoint () {
    const s = this.layer.viewport.scale
    const startX = this.start.x
    const startY = this.start.y
    const endX = this.end.x
    const endY = this.end.y

    let x1, y1, x2, y2;
    x1 = startX / s;
    y1 = startY / s; 
    x2 = endX / s;
    y2 = endY / s;
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    };
  }

  selectAnnotation () {
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) return
    this.handleClick()
  }

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
        this.layer.annotationStore.selectedElementName = null
      }
      this.outerLineContainer.remove()
    } else {
      if (this.layer.annotationStore.selectedElementName !== this.annotation.name) {
        const annotation = this.annotation
        this.layer.annotationStore.selectedElementName = annotation.name

        const annotationData = {
          ...annotation
        }
        delete annotationData.annotPtr
        delete annotationData.pagePtr
        delete annotationData.operate
        delete annotationData.targetPage

        this.layer.eventBus.dispatch('annotationSelected', annotationData)
      }
      this.layer.div.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)

      if (this.annotation.measure) {
        this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
        this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
        this.layer.eventBus.dispatch('updateMeasureData', {
          scale: this.annotation.scale,
          precision: this.annotation.precision,
          perimeter: this.annotation.perimeter,
          area: this.annotation.area,
          showPerimeter: this.showPerimeter,
          showArea: this.showArea
        })
        this.showPopup()
      }

      const documentViewer = this.layer.documentViewer
      const scrollElement = documentViewer.getScrollViewElement()
      scrollElement.style.touchAction = 'none'
    }
  }

  handleClick (event) {
    if (!this.hidden || this.layer.annotationStore.creating || ['redaction', 'remove'].includes(this.layer.tool) || document.querySelector('.compare-document-container').contains(event?.target)) return
    this.hidden = false
    this.updateTool()

    if (this.annotation.measure) {
      this.layer.eventBus._on('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._on('measurementPopupClicked', this.onHandlePopup)
    }
    document.addEventListener('keydown', this.onKeydown)

    const ignoreList = [
      this.svgElement,
      this.outerLine,
      document.querySelector('.toggle-button'),
      document.getElementById('propertyPanelButton'),
      document.querySelector('.measure-panel'),
      document.querySelector('.measurement-board'),
      document.querySelector('.measurements-settings'),
      document.querySelector('.measurement-popup')
    ]
    onClickOutsideUp(ignoreList, this.handleOutside.bind(this))
  }

  handleOutside () {
    if (!this.hidden && this.annotation.measure && this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
      this.layer.eventBus.dispatch('measurePropertyChange', { isOpen: false })
      this.hidePopup()
    }

    this.hidden = true
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.annotationStore.selectedElementName = null
    }

    document.removeEventListener('keydown', this.onKeydown)
    this.outerLine?.removeEventListener('mousedown', this.onMousedown)
    this.outerLine?.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer?.remove()

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    scrollElement.style.touchAction = 'auto'

    if (this.annotation.measure) {
      this.layer.eventBus._off('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._off('measurementPopupClicked', this.onHandlePopup)
    }
  }

  handleMouseDown (event) {
    if (event.button !== 0 && event.type === 'mousedown') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    this.scrollTop = scrollElement.scrollTop
    this.scrollLeft = scrollElement.scrollLeft

    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  };

  handleMouseMove (event) {
    if (event.button !== 0 && event.type === 'mousemove') return
    this.hidePopup()

    if (event.type === 'touchmove') {
      this.layer.pageDiv.style.touchAction = 'none'
    }
    this.moving = true
    let { pageX, pageY } = getClickPoint(event)

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    const scrollTop = scrollElement.scrollTop
    const scrollLeft = scrollElement.scrollLeft
    pageX = pageX + scrollLeft - this.scrollLeft
    pageY = pageY + scrollTop - this.scrollTop

    const start = this.start
    const end = this.end
    const startState = this.startState

    const { width, height } = this.layer.viewport
    if (startState.operator === 'nw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(end.y - 5, y1)
      y1 = Math.max(0, y1)

      this.update({
        start: { x: x1, y: y1 },
        end,
      });
    } else if (startState.operator === 'w-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)

      this.update({
        start: { x: x1, y: start.y },
        end,
      });
    } else if (startState.operator === 'sw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.end.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(start.y + 5, y1)

      this.update({
        start: { x: x1, y: start.y },
        end: { x: end.x, y: y1 },
      });
    } else if (startState.operator === 's-resize') {
      let y1 = pageY - (startState.clickY - this.end.y);
      y1 = Math.min(height, y1)
      y1 = Math.max(start.y + 5, y1)
      this.update({
        start,
        end: { x: end.x, y: y1 },
      });
    } else if (startState.operator === 'se-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.end.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(height, y)
      y = Math.max(start.y + 5, y)
      this.update({
        start,
        end: { x, y },
      });
    } else if (startState.operator === 'e-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      this.update({
        start,
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'ne-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.start.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(end.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start: { x: start.x, y },
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'n-resize') {
      let y = pageY - (startState.clickY - this.start.y);
      y = Math.min(end.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start: { x: start.x, y },
        end,
      });
    } else if (startState.operator === 'move') {
      let x1, y1, x2, y2
      const documentViewer = this.layer.documentViewer

      let windowCoordinates = this.getMouseLocation(event);
      const page = documentViewer.getSelectedPage(windowCoordinates, windowCoordinates)
      if (page.first && page.first !== (this.layer.page + 1)) {
        const pagePoint = documentViewer.windowToPage(windowCoordinates, page.first, false)

        const annotationWidth = Math.abs(start.x - end.x)
        const annotationHeight = Math.abs(start.y - end.y)
        if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 0) {
          x1 = pageX - (startState.clickX - this.start.x)
          y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
          x2 = pageX - (startState.clickX - this.end.x)
          y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
        } else if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 1) {
          x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
          y1 = pageY - (startState.clickY - this.start.y)
          x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
          y2 = pageY - (startState.clickY - this.end.y)
        } else if (documentViewer.pdfViewer.spreadMode === 1 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && minPage % 2 === 0)) {
            return
          }
          if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          } else if ((minPage + 2) === maxPage) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          }
        } else if (documentViewer.pdfViewer.spreadMode === 2 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && !(minPage === 1 && maxPage === 2) && maxPage % 2 === 0 || (minPage === 1 && maxPage === 3))) {
            return
          }
          if ((minPage + 2) === maxPage || (minPage === 1 && maxPage === 2)) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          } else if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          }
        }
        this.layer = documentViewer.pdfViewer._pages[page.first - 1].compdfAnnotationLayer

        this.annotationContainer.remove()
        this.outerLineContainer.remove()
        const container = this.layer.div
        container.append(this.annotationContainer)
        container.append(this.outerLineContainer)
      } else {
        y1 = pageY - (startState.clickY - this.start.y);
        x1 = pageX - (startState.clickX - this.start.x);
        x2 = pageX - (startState.clickX - this.end.x);
        y2 = pageY - (startState.clickY - this.end.y);
        const windowPoint1 = documentViewer.pageToWindow({
          x: x1,
          y: y1
        }, this.layer.page + 1, false)
        const windowPoint2 = documentViewer.pageToWindow({
          x: x2,
          y: y2
        }, this.layer.page + 1, false)
        let x = 0, y = 0
        if (pageY > startState.clickY) {
          y = scrollElement.offsetHeight - (windowPoint2.y - scrollElement.scrollTop - scrollElement.offsetTop)
          y = y < 100 ? 5 : 0
        } else if (pageY < startState.clickY) {
          y = windowPoint1.y - scrollElement.scrollTop - scrollElement.offsetTop
          y = y < 100 ? -5 : 0
        }
        if (pageX > startState.clickX) {
          x = scrollElement.offsetWidth - (windowPoint2.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? 5 : 0
        } else if (pageX < startState.clickX) {
          x = (windowPoint1.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? -5 : 0
        }

        if (x || y) {
          scrollElement.scrollTop += y
          scrollElement.scrollLeft += x
        }
        const rect = {
          width: Math.abs(start.x - end.x),
          height: Math.abs(start.y - end.y)
        }
        if (x1 < x2) {
          x1 = Math.max(0, x1)
          x1 = Math.min(width - rect.width, x1)
          x2 = Math.max(rect.width, x2)
          x2 = Math.min(width, x2)
        } else {
          x2 = Math.max(0, x2)
          x2 = Math.min(width - rect.width, x2)
          x1 = Math.max(rect.width, x1)
          x1 = Math.min(width, x1)
        }

        if (y1 < y2) {
          y1 = Math.max(0, y1)
          y1 = Math.min(height - rect.height, y1)
          y2 = Math.max(rect.height, y2)
          y2 = Math.min(height, y2)
        } else {
          y2 = Math.max(0, y2)
          y2 = Math.min(height - rect.height, y2)
          y1 = Math.max(rect.height, y1)
          y1 = Math.min(height, y1)
        }
      }

      this.update({
        start: { x: x1, y: y1 },
        end: { x: x2, y: y2 },
      });
    }
    startState.clickX = pageX
    startState.clickY = pageY
  }

  handleDelete (event) {
    if (this.layer.tool && event) {
      event.stopPropagation()
    }

    this.delete()
    const annotationData = {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        pageIndex: this.layer.page
      }
    }
    this.annotation.annotPtr && (annotationData.annotation.annotPtr = this.annotation.annotPtr)

    this.layer.eventBus.dispatch('annotationChanged', annotationData)
  }

  delete () {
    this.handleOutside()
    this.annotationContainer?.remove()
    this.annotation.isDelete = true
    this.layer.deleteAnnotationInLayer({
      name: this.annotation.name
    })
  }

  async handleMouseUp (event) {
    if (event.button !== 0 && event.type === 'mouseup') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      this.layer.pageDiv.style.touchAction = ''
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    this.scrollLeft = 0
    this.scrollTop = 0

    const annotation = this.annotation
    const { start, end } = this.getInitialPoint()

    const rect = {
      left: start.x,
      top: start.y,
      right: end.x,
      bottom: end.y
    }
    const newRect = this.transformRect(rect)
    const vertices = [
      { x: newRect.left, y: newRect.top },
      { x: newRect.right, y: newRect.top },
      { x: newRect.right, y: newRect.bottom },
      { x: newRect.left, y: newRect.bottom }
    ]

    if (rect === annotation.rect && annotation.pageIndex === this.layer.page) return

    annotation.rect = rect
    annotation.vertices = vertices
    this.annotation.measure && this.showPopup()

    if (annotation.pageIndex !== this.layer.page) {
      const annotationData = {
        type: 'delete',
        annotation: {
          operate: "del-annot",
          name: annotation.name,
          pageIndex: annotation.pageIndex,
          annotPtr: annotation.annotPtr
        }
      }
      this.layer.eventBus.dispatch('annotationChanged', annotationData)

      const newAnnotationData = Object.assign({}, annotation)
      delete newAnnotationData.annotPtr
      delete newAnnotationData.pagePtr
      delete newAnnotationData.name
      newAnnotationData.operate = 'add-annot'
      newAnnotationData.pageIndex = this.layer.page
      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'add',
      //   show: true,
      //   annotation: newAnnotationData
      // })
      const annots = await this.layer.documentViewer.annotationManager.createMeasurementAnnotation(newAnnotationData)
      this.annotation = annots[0]
      await this.layer.documentViewer.annotationManager.drawAnnotationsFromList(this.annotation)

      this.annotationContainer.remove()
      this.outerLineContainer.remove()
    } else {
      const annotationData = {
        operate: "mod-annot",
        name: annotation.name,
        pageIndex: this.layer.page,
        pagePtr: annotation.pagePtr,
        annotPtr: annotation.annotPtr,
        rect,
        vertices
      }

      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'modify',
      //   annotation: annotationData
      // })
      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
    }
  }

  update ({ start, end }, isUpdateMeasure = true) {
    const rect = this.calculate(start, end)
    this.rect = rect
    this.start = start
    this.end = end

    const actualbdwidth = (this.annotation.borderWidth || 2) * this.layer.viewport.scale

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'
    
    this.svgElement.setAttribute("viewBox", `0 0 ${rect.width} ${rect.height}`)

    this.shapeElement.setAttribute("x", actualbdwidth / 2)
    this.shapeElement.setAttribute("y", actualbdwidth / 2)
    this.shapeElement.setAttribute("width", rect.width - actualbdwidth)
    this.shapeElement.setAttribute("height", rect.height - actualbdwidth)

    this.outerLine.style.left = `${rect.left - 7}px`
    this.outerLine.style.top = `${rect.top - 7}px`
    this.outerLine.style.width = `${rect.width + 7 * 2}px`
    this.outerLine.style.height = `${rect.height + 7 * 2}px`
    
    this.moveRect.setAttribute('width',  rect.width + 8)
    this.moveRect.setAttribute('height',  rect.height + 8)
    
    this.bottomLeftRect.setAttribute("y", rect.height + 8)

    this.bottomRightRect.setAttribute("x", rect.width + 8)
    this.bottomRightRect.setAttribute("y", rect.height + 8)
    
    this.topRightRect.setAttribute("x", rect.width + 8)

    this.annotation.measure && isUpdateMeasure && this.updateMeasure(start, end)
  }

  openReply () {
    this.layer.eventBus.dispatch('openAnnotationReply', this.annotation)
  }

  async updateMeasure (start, end) {
    // 更新测量数据
    if (this.startState && this.startState.operator !== 'move') {
      const initPoint = this.getInitialPoint()
      const rect = {
        left: Math.min(initPoint.start.x, initPoint.end.x),
        top: Math.min(initPoint.start.y, initPoint.end.y),
        right: Math.max(initPoint.start.x, initPoint.end.x),
        bottom: Math.max(initPoint.start.y, initPoint.end.y)
      }
      const newRect = this.transformRect(rect)
      const vertices = [
        { x: newRect.left, y: newRect.top },
        { x: newRect.right, y: newRect.top },
        { x: newRect.right, y: newRect.bottom },
        { x: newRect.left, y: newRect.bottom }
      ]

      const annotationData = {
        pagePtr: this.annotation.pagePtr,
        annotPtr: this.annotation.annotPtr,
        rulerPtr: this.annotation.rulerPtr,
        pageIndex: this.annotation.pageIndex,
        rect,
        vertices,
      }
      if (this.annotation.contents) {
        annotationData.contents = this.annotation.contents
        annotationData.labelPosition = this.annotation.labelPosition
      }
      this.annotation.rect = rect
      this.annotation.vertices = vertices

      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
      await this.getMeasureData()
      const textList = []
      this.showPerimeter && textList.push(this.annotation.perimeter + this.annotation.scale.translateUnit)
      this.showArea && textList.push(this.annotation.area + this.annotation.scale.translateUnit + '\u00B2')
      this.annotation.contents = textList.join('\n')
    }

    // 更新绘制的元素
    if (this.permiterTspan) {
      this.permiterTspan.setAttribute('x', (end.x - start.x) / 2)
      this.permiterTspan.setAttribute('y', (end.y - start.y) / 2)
      this.permiterTspan.textContent = this.annotation.perimeter + this.annotation.scale.translateUnit
    }
    if (this.areaTspan) {
      this.areaTspan.setAttribute('x', (end.x - start.x) / 2)
      this.areaTspan.setAttribute('y', (end.y - start.y) / 2)
      this.areaTspan.textContent = this.annotation.area + this.annotation.scale.translateUnit + '\u00B2'
    }

    if (this.startState && this.startState.operator === 'move') return

    // 更新看板数据
    this.layer.eventBus.dispatch('updateMeasureData', {
      scale: this.annotation.scale,
      precision: this.annotation.precision,
      perimeter: this.annotation.perimeter,
      area: this.annotation.area
    })
  }

  async handlePropertyPanel(props) {
    if (this.layer.annotationStore.selectedElementName !== this.annotation.name) return

    let updatePropsNum = 0

    for (const key in props) {
      updatePropsNum += await this.setCustomData(key, props[key], true)
    }

    updatePropsNum > 0 && this.sendData(props)
  }

  async setCustomData(key, value, noSend = false) {
    const annotation = this.annotation
    let isUpdate = 0

    if (key === 'showPerimeter' && this.showPerimeter !== value) {
      this.showPerimeter = value
      if (value && !this.permiterTspan) {
        this.permiterTspan = createSvg('tspan', {
          x: (this.end.x - this.start.x) / 2,
          y: (this.end.y - this.start.y) / 2,
          dy: '0em'
        })
        this.permiterTspan.textContent = annotation.perimeter + annotation.scale.translateUnit
        this.textEle.appendChild(this.permiterTspan)
      } else if (this.permiterTspan) {
        this.permiterTspan.remove()
        this.permiterTspan = null
      }
      return 0
    }
    if (key === 'showArea' && this.showArea !== value) {
      this.showArea = value
      if (value && !this.areaTspan) {
        this.areaTspan = createSvg('tspan', {
          x: (this.end.x - this.start.x) / 2,
          y: (this.end.y - this.start.y) / 2,
          dy: '1em'
        })
        this.areaTspan.textContent = annotation.area + annotation.scale.translateUnit + '\u00B2'
        this.textEle.appendChild(this.areaTspan)
      } else if (this.areaTspan) {
        this.areaTspan.remove()
        this.areaTspan = null
      }
      return 0
    }

    for (const item in annotation) {
      if (item === key && annotation[item] !== value) {
        annotation[item] = value
        isUpdate = 1

        // Styles
        if (key === 'borderColor') {
          this.shapeElement.setAttribute('stroke', value)
        }
        if (key === 'fillColor') {
          this.shapeElement.setAttribute('fill', value)
        }
        if (key === 'opacity') {
          this.svgElement.style.opacity = value
          annotation.fillTransparency = value
        }
        if (key === 'borderWidth') {
          this.shapeElement.setAttribute('stroke-width', value)
          this.update({ start: this.start, end: this.end }, false)
          this.showPopup()
        }
        if (key === 'borderStyle') {
          if (value === 'solid') {
            this.shapeElement.setAttribute('stroke-dasharray', '')
          }
          if (value === 'dashed') {
            let dashString = ''
            const scaleValue = Math.ceil(this.ratio * this.layer.viewport.scale)

            if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
              dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
            } else if (annotation.dashes.length > 2) {
              dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
            }
            this.shapeElement.setAttribute('stroke-dasharray', dashString)
          }
        }
        if (key === 'dashes') {
          this.shapeElement.setAttribute('stroke-dasharray', Math.ceil(value[0] * this.ratio * this.layer.viewport.scale))
        }
        if (key === 'color') {
          this.textEle.setAttribute('fill', value)
        }
        if (key === 'fontSize') {
          this.textEle.setAttribute('font-size', value)
        }
        if (key === 'fontName') {
          this.textEle.style.fontFamily = value
        }
        if (key === 'fontStyle') {
          const fontStyle = value
          if (fontStyle === 'boldItalic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = true
          } else if (fontStyle === 'bold') {
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = false
          } else if (fontStyle === 'italic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'normal'
            annotation.isBold = false
            annotation.isItalic = true
          } else {
            annotation.isItalic = false
            annotation.isBold = false
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'normal'
          }
        }
        if (key === 'contents') {
          if (this.areaTspan) {
            this.areaTspan.remove()
            this.areaTspan = null
          }
          this.permiterTspan.textContent = value
        }

        // Measure Properties
        if (key === 'scale' || key === 'precision') {
          if (key === 'precision') {
            this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
          }

          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            rulerPtr: annotation.rulerPtr,
            [key]: value
          })
          await this.getMeasureData()
          const textList = []
          if (this.showPerimeter && this.permiterTspan) {
            const text = annotation.perimeter + annotation.scale.translateUnit
            textList.push(text)
            this.permiterTspan.textContent = text
          }
          if (this.showArea && this.areaTspan) {
            const text = annotation.area + annotation.scale.translateUnit + '\u00B2'
            textList.push(text)
            this.areaTspan.textContent = text
          }
          annotation.contents = textList.join('\n')

          if (this.layer.annotationStore.selectedElementName === this.annotation.name && !document.querySelector('.measure-panel').classList.contains('closed')) {
            this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
          }

          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            annotPtr: annotation.annotPtr,
            contents: annotation.contents
          })

          this.layer.eventBus.dispatch('updateMeasureData', {
            scale: annotation.scale,
            precision: annotation.precision,
            perimeter: annotation.perimeter,
            area: annotation.area
          })

          return 0
        }
      }
    }
    !noSend && this.sendData({ [key]: value })
    return isUpdate
  }

  sendData(props) {
    const annotation = this.annotation
    const annotationData = {
      type: 'polygon',
      pagePtr: annotation.pagePtr,
      annotPtr: annotation.annotPtr,
      rulerPtr: annotation.rulerPtr,
      pageIndex: annotation.pageIndex,
      contents: annotation.contents
    }

    const isFont = ['fontSize', 'fontName', 'color', 'fontStyle'].some(item => props[item])
    if (isFont) {
      annotationData.fontSize = annotation.fontSize
      annotationData.color = annotation.color
      const fontArr = {
        Courier: {
          normal: 'Courier',
          bold: 'Courier-Bold',
          italic: 'Courier-Oblique',
          boldItalic: 'Courier-BoldOblique'
        },
        Helvetica: {
          normal: 'Helvetica',
          bold: 'Helvetica-Bold',
          italic: 'Helvetica-Oblique',
          boldItalic: 'Helvetica-BoldOblique'
        },
        Times: {
          normal: 'Times',
          bold: 'Times-Bold',
          italic: 'Times-Italic',
          boldItalic: 'Times-BoldItalic'
        }
      }

      const fontName = annotation.fontName
      if (annotation.isBold && annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].boldItalic
      } else if (annotation.isBold) {
        annotationData.fontName = fontArr[fontName].bold
      } else if (annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].italic
      } else {
        annotationData.fontName = fontArr[fontName] ? fontArr[fontName].normal : fontName
      }
      delete props.fontName
    }
    if (props.dashes) {
      if (props.dashes.length === 1) {
        props.dashes = new Array(2).fill(props.dashes[0])
      }
      annotationData.borderStyle = annotation.borderStyle
      annotationData.borderWidth = annotation.borderWidth
    }
    if (props.borderStyle) {
      annotationData.dashes = props.borderStyle === 'solid' ? [] : annotation.dashes.length === 1 ? new Array(2).fill(annotation.dashes[0]) : annotation.dashes
      annotationData.borderWidth = annotation.borderWidth
    }
    if (props.opacity) {
      annotationData.fillTransparency = annotation.fillTransparency
    }
    if (props.fillColor) {
      annotationData.fillTransparency = annotation.fillTransparency
    }

    const updateAnnot = Object.assign({}, annotationData, props)
    this.layer.documentViewer.annotationManager.editMeasurementAnnotation(updateAnnot)
  }

  async getMeasureData() {
    const { rulerPtr } = this.annotation
    const area = await this.layer.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { rulerPtr, captionType: 'area' })
    const perimeter = await this.layer.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { rulerPtr, captionType: 'length' })

    this.annotation.area = this.formatValue(area)
    this.annotation.perimeter = this.formatValue(perimeter)
  }

  showPopup() {
    const pageRect = this.layer.div.getBoundingClientRect()

    const left = pageRect.left + this.rect.left - 7
    const top = this.rect.top + pageRect.top - 7
    const right = left + this.rect.width + 7 * 2
    const bottom = top + this.rect.height + 7 * 2
    const rect = {
      left,
      top,
      right,
      bottom
    }
    this.layer.eventBus.dispatch('showMeasurementPopup', { rect })
    this.isPopupShow = true
  }

  hidePopup() {
    if (!this.isPopupShow) return
    this.isPopupShow = false
    this.layer.eventBus.dispatch('showMeasurementPopup')
  }

  async handlePopup(data) {
    if (data === 'delete') {
      this.handleDelete()
    }
    if (data === 'reply') {
      this.openReply()
    }
  }

  // 转为以左下角为原点
  transformRect(rect) {
    const { viewBox, rotation } = this.layer.viewport
    const width = viewBox[2]
    const height = viewBox[3]

    const radians = rotation * Math.PI / 180;

    // 中心点
    const centerX = width / 2;
    const centerY = height / 2;

    // 提取左上角和右下角的坐标
    const startX = rect.left;
    const startY = rect.top;
    const endX = rect.right;
    const endY = rect.bottom;

    // 旋转矩阵
    const cosTheta = Math.cos(radians);
    const sinTheta = Math.sin(radians);

    // 旋转函数
    const rotatePoint = (x, y) => {
      const rotatedX = cosTheta * (x - centerX) - sinTheta * (y - centerY) + centerX;
      const rotatedY = sinTheta * (x - centerX) + cosTheta * (y - centerY) + centerY;
      return { x: rotatedX, y: rotatedY };
    }

    // 旋转左上角和右下角点
    const rotatedStart = rotatePoint(startX, startY);
    const rotatedEnd = rotatePoint(endX, endY);

    // 转换为以左下角为原点
    const left = rotatedStart.x;
    const top = height - rotatedStart.y;
    const right = rotatedEnd.x;
    const bottom = height - rotatedEnd.y;

    return { left, top, right, bottom };
  }

  formatValue(value) {
    return value % 1 === 0 ? value : Number(value.toFixed(this.decimalPlaces))
  }

  async handleKeyDown (e) {
    if (e.target.tagName === 'INPUT' || e.target.tagName === 'TEXTAREA') return

    let keyCode = e.keyCode || e.which
    if (keyCode === 8 || keyCode === 46) { // 8 backspace键，46 delete键
      this.handleDelete()
    }
  }
}
