import {
  getAbsoluteCoordinate,
  createElement,
  getInitialPoint
} from '../annotation/utils';

export default class AddPushButton {
  constructor ({
    tool,
    color,
    layer,
  }) {
    this.layer = layer
    this._tool = tool
    this.color = color
    this.path = []

    this.initStartPoint = null
    this.initEndPoint = null
    this.buttonEle = null

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.initPushButtonName = this.setInitProperty.bind(this)

    this.init()
  }

  init () {
    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)

    this.layer.pageDiv.addEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.addEventListener('touchstart', this.onMousedown)

    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)

    this.layer.eventBus._on('initPushButtonName', this.initPushButtonName)
  }

  get tool () {
    return this._tool
  }

  set tool (toolType) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
      this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
      this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)
    }
    this._tool = toolType
  }

  reset () {
    this.initStartPoint = null
    this.initEndPoint = null
    if (this.buttonEle) this.buttonEle.remove()
    this.buttonEle = null
    this.layer.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)

    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    this.layer.pageDiv.removeEventListener('touchstart', this.onMousedown)

    this.layer.eventBus._off('initPushButtonName', this.initPushButtonName)
  }

  handleMouseDown (event) {
    if (this.layer.annotationStore.selectedElementName) return
    const { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    this.initStartPoint = { x, y }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  }

  handleMouseUp () {
    let initStartPoint = this.initStartPoint
    let initEndPoint = this.initEndPoint
    if (initEndPoint && initStartPoint.x !== initEndPoint.x && initStartPoint.y !== initEndPoint.y) {
      let start = getInitialPoint(initStartPoint, this.layer.viewport.scale)
      let end = getInitialPoint(initEndPoint, this.layer.viewport.scale)
      
      this.setInitProperty()

      const newStart = {
        x: Math.min(start.x, end.x),
        y: Math.min(start.y, end.y)
      }
      const newEnd = {
        x: Math.max(start.x, end.x),
        y: Math.max(start.y, end.y)
      }
      
      const rect = {
        left: newStart.x,
        top: newStart.y,
        right: newEnd.x,
        bottom: newEnd.y
      }

      const annotationData = {
        operate: "add-form",
        type: 'pushbutton',
        pageIndex: this.layer.page,
        date: new Date(),
        rect,
        backgroundColor: '#DDE9FF',
        borderColor: '#43474D',
        borderStyle: 'solid',
        borderWidth: 1,
        fieldName: this.fieldName,
        isHidden: 0,

        color: '#000000',
        fontName: 'Helvetica',
        fontSize: 14,
        title: 'OK',
        actionType: 0,
        destPage: '',
        url: ''
      }

      this.layer.eventBus.dispatch('annotationChanged', {
        type: 'add',
        show: true,
        annotation: annotationData
      })
    }
    // this.reset()
    
    this.initStartPoint = null
    this.initEndPoint = null
    if (this.buttonEle) this.buttonEle.remove()
    this.buttonEle = null
    this.layer.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
  }

  handleMouseMove (event) {
    let { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    const { width, height } = this.layer.viewport
    const scale = this.layer.viewport.scale
    x = Math.min(width, x)
    x = Math.max(0, x)
    y = Math.min(height, y)
    y = Math.max(0, y)
    this.initEndPoint = { x, y }
    if (this.initStartPoint.x !== this.initEndPoint.x && this.initStartPoint.y !== this.initEndPoint.y) {
      if (!this.buttonEle) {
        const buttonEle = createElement('div', {
          'display': 'flex',
          'alignItems': 'center',
          'justifyContent': 'center',
          'color': '#000000',
          'fontSize': '14px',
          'borderColor': '#001A4E',
          'borderWidth': '1px',
          'borderStyle': 'solid',
          'backgroundColor': '#DDE9FF',
          'position': 'absolute',
          'fontFamily': 'Helvetica',
          'overflow': 'hidden'
        })
        buttonEle.innerHTML = 'OK'
        this.buttonEle = buttonEle
        this.scene = this.layer.pageDiv.querySelector('.annotationLayer')
        this.scene.append(this.buttonEle)
      } else {
        if (event.shiftKey) {
          const { initStartPoint, initEndPoint } = this
          const distanceX = Math.abs(initEndPoint.x - initStartPoint.x)
          const distanceY = Math.abs(initEndPoint.y - initStartPoint.y)
          const distance = Math.min(distanceX, distanceY)
          if (initStartPoint.x < initEndPoint.x) {
            initEndPoint.x = initStartPoint.x + distance
          } else {
            initEndPoint.x = initStartPoint.x - distance
          }
          if (initStartPoint.y < initEndPoint.y) {
            initEndPoint.y = initStartPoint.y + distance
          } else {
            initEndPoint.y = initStartPoint.y - distance
          }
        }
        const rect = this.calculate(this.initStartPoint, this.initEndPoint)

        const buttonEle = this.buttonEle
        buttonEle.style.top = `${rect.top}px`;
        buttonEle.style.left = `${rect.left}px`;
        buttonEle.style.width = `${Math.abs(rect.width - 1)}px`;
        buttonEle.style.height = `${Math.abs(rect.height - 1)}px`;
      }
    }
  }
  
  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = 1 * this.layer.viewport.scale

    return {
      top: initRect.top - actualbdwidth,
      left: initRect.left - actualbdwidth,
      width: initRect.width + actualbdwidth * 2,
      height: initRect.height + actualbdwidth * 2,
    }
  }

  setInitProperty () {
    let index = 1
    if (this.layer.annotationStore.annotationsAll) {
      let annotArr = Object.values(this.layer.annotationStore.annotationsAll).reduce(function(acc, curr) {
        return acc.concat(curr);
      }, []);

      const textfieldArr = annotArr.filter(item => item.type === 'pushbutton')
      for (let i = 0; i <= textfieldArr.length; i++) {
        const name = 'Button' + (i + 1)
        if (textfieldArr.every(item => item.fieldName !== name)) {
          index = i + 1
          break
        }
      }
    }
    this.fieldName = 'Button' + index
  }
}
