/* Copyright 2012 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { createPromiseCapability, PasswordResponses } from "pdfjs-dist/legacy/build/pdf";

/**
 * @typedef {Object} PasswordPromptOptions
 * @property {HTMLDialogElement} dialog - The overlay's DOM element.
 * @property {HTMLParagraphElement} label - Label containing instructions for
 *                                          entering the password.
 * @property {HTMLInputElement} input - Input field for entering the password.
 * @property {HTMLButtonElement} submitButton - Button for submitting the
 *                                              password.
 * @property {HTMLButtonElement} cancelButton - Button for cancelling password
 *                                              entry.
 */

class PasswordPrompt {
  #activeCapability = null;

  #updateCallback = null;

  #callback = null;

  #reason = null;

  #active = false

  /**
   * @param {PasswordPromptOptions} options
   * @param {OverlayManager} overlayManager - Manager for the viewer overlays.
   * @param {IL10n} l10n - Localization service.
   *   an <iframe> or an <object>. The default value is `false`.
   */
  constructor(options) {
    this.dialog = options.dialog;
    this.label = options.label;
    this.input = options.input;
    this.submitButton = options.submitButton;
    this.cancelButton = options.cancelButton;
    this.$t = options.$t;

    // Attach the event listeners.
    this.submitButton.addEventListener("click", this.#verify.bind(this));
    this.cancelButton.addEventListener("click", this.close.bind(this));
    this.input.addEventListener("keyup", e => {
      if (e.keyCode === /* Enter = */ 13) {
        this.#verify();
      }

      this.label.textContent = ''

      if (this.input.value === '') {
        this.submitButton.classList.add('disabled')
      } else {
        this.submitButton.classList.remove('disabled')
      }
    });
  }

  async open() {
    if (this.#activeCapability) {
      await this.#activeCapability.promise;
    }
    this.#activeCapability = createPromiseCapability();

    try {
      this.dialog.classList.add('show');
      this.submitButton.classList.add('disabled')
      this.#active = true;
    } catch (ex) {
      this.#activeCapability = null;
      throw ex;
    }

    const passwordIncorrect =
      this.#reason === PasswordResponses.INCORRECT_PASSWORD;

    if (passwordIncorrect) {
      this.input.focus();
    }
    if (passwordIncorrect) {
      this.label.textContent = this.$t('passwordDialog.wrongPwd')
      this.label.style.color = 'red'
    } else {
      this.label.style.color = ''
      this.label.textContent = ''
    }
  }

  async close() {
    if (this.#active) {
      const event = new Event('click');
      event.param = false;
      document.getElementById('eyeVisible').dispatchEvent(event);

      this.dialog.classList.remove('show');
      this.#active = false;
      this.#cancel()
    }
  }

  #verify() {
    const password = this.input.value;
    if (password?.length > 0) {
      this.#invokeCallback(password);
    }
  }

  #cancel() {
    this.#invokeCallback(new Error("PasswordPrompt cancelled."));
    this.#activeCapability.resolve();
  }

  #invokeCallback(password) {
    if (!this.#updateCallback) {
      return; // Ensure that the callback is only invoked once.
    }
    this.input.value = "";

    this.#updateCallback(password);
    this.#callback(password);
    this.#callback = null;
    this.#updateCallback = null;
    this.close();
  }

  async setUpdateCallback(updateCallback, reason, callback) {
    if (this.#activeCapability) {
      await this.#activeCapability.promise;
    }
    this.#updateCallback = updateCallback;
    this.#reason = reason;
    this.#callback = callback
  }
}

export { PasswordPrompt };
