// @ts-nocheck
import ArrowHelper from '../annotation/arrow'
import { getAbsoluteCoordinate, createSvg, getClickPoint, getInitialPoint } from '../annotation/utils'

interface Point {
  x: number,
  y: number
}

export default class AreaMeasurementCreateTool {
  documentViewer: any
  tool: any
  eventBus: any

  measureType: string
  defaults: {
    scale: {
      base: number
      baseUnit: string
      translate: number
      translateUnit: string
    }
    precision: number
    factor: number
  }
  defaultStyles: {
    borderWidth: number
    opacity: number
    fillTransparency: number
    color: string
    borderColor: string
    borderStyle: string
    dashes: number[]
    fontName: string
    fontSize: number
    labelPosition: string
  }
  drawingPageIndex: number = -1 // 绘制的页面索引，大于等于0代表正在绘制中
  drawingLayer: any = null
  realStartPoint: Point | null = null
  realEndPoint: Point | null = null
  oriStartPoint: Point | null = null
  oriEndPoint: Point | null = null
  shapeEle: HTMLDivElement | null = null
  textEle: HTMLDivElement | null = null
  annotation: any | null = null // 正在创建的注释信息
  decimalPlaces: number
  showPerimeter: boolean
  showArea: boolean

  onMousedown = this.handleMouseDown.bind(this)
  onMousemove = this.handleMouseMove.bind(this)
  onMouseup = this.handleMouseUp.bind(this)
  handleTool = this.handleTool.bind(this)

  constructor(options: {
    documentViewer: any,
    annotationStore: any
  }) {
    this.documentViewer = options.documentViewer
    this.tool = options.documentViewer.activeTool
    this.eventBus = options.documentViewer.eventBus
    this.annotationStore = options.annotationStore

    this.measureType = 'area'
    this.defaults = { // 测量属性
      scale: {
        base: 1,
        baseUnit: 'cm',
        translate: 1,
        translateUnit: 'cm',
      },
      precision: 1,
    }
    this.defaultStyles = { // 注释属性
      borderWidth: 2,
      opacity: 1,
      fillTransparency: 1,
      color: '#FF0000',
      borderColor: '#FF0000',
      borderStyle: 'solid',
      dashes: [],
      // fillColor: '#FF0000',
      fontName: 'Helvetica', // Helvetica, Helvetica-Bold, Helvetica-BoldOblique, Courier, Courier-Oblique, Times, Times-Italic
      fontSize: 14,
      fontStyle: 'regular',
      // contents: 'test',
      labelPosition: 'inside', // inside / top
    }
    this.decimalPlaces = this.defaults.precision.toString().split('.')[1]?.length || 0 // 保留小数位
    this.showPerimeter = true
    this.showArea = true

    this.eventBus._on('toolChanged', this.handleTool)
  }

  handleTool({
    tool,
    color
  }) {
    const oldtool = this.tool
    this.tool = tool

    if (oldtool === 'rectangle' && this.tool !== oldtool) {
      this.reset()
    }
    if (this.tool === 'rectangle' && this.documentViewer.toolMode === 'measurement') {
      this.init()
    }
  }

  init() {
    document.removeEventListener('mousedown', this.onMousedown)
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)

    document.addEventListener('mousedown', this.onMousedown)

    this.documentViewer.viewerContainer.classList.add('annotation-edit')
  }

  reset() {
    this.realStartPoint = null
    this.realEndPoint = null
    this.oriStartPoint = null
    this.oriEndPoint = null
    this.shapeEle = null
    this.textEle = null

    document.removeEventListener('mousedown', this.onMousedown)
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)

    this.documentViewer.viewerContainer.classList.remove('annotation-edit')
  }

  async handleMouseDown(event) {
    if (event.target.className !== 'annotationContainer') return

    const windowCoordinates = this.getMouseLocation(event)
    const page = this.documentViewer.getSelectedPage(windowCoordinates, windowCoordinates)
    const pageIndex = page.first - 1
    if (pageIndex < 0) return
    if (this.annotationStore.selectedElementName && this.drawingPageIndex < 0) return

    const layer = this.documentViewer.pdfViewer._pages[pageIndex].compdfAnnotationLayer
    const realPoint = getAbsoluteCoordinate(layer.pageDiv, event) // 实际坐标
    const oriPoint = this.documentViewer.windowToPage(windowCoordinates, pageIndex + 1) // 原始坐标

    this.drawingPageIndex = pageIndex
    this.drawingLayer = layer
    this.realStartPoint = { x: realPoint.x, y: realPoint.y }
    this.oriStartPoint = { x: oriPoint.x, y: oriPoint.y }

    const start = this.oriStartPoint
    // const rect = {
    //   left: start.x,
    //   top: start.y,
    //   right: start.x + 1,
    //   bottom: start.y + 1
    // }
    // const vertices = [
    //   { x: rect.left, y: rect.top },
    //   { x: rect.right, y: rect.top },
    //   { x: rect.right, y: rect.bottom },
    //   { x: rect.left, y: rect.bottom }
    // ]
    let annotationData: Record<string, unknown> = {
      measure: 1,
      operate: 'add-annot',
      type: 'polygon',
      measureType: this.measureType,
      pageIndex,
      date: new Date(),
      // rect,
      // vertices,
      ...this.defaults,
      ...this.defaultStyles,
    }
    const annots = await this.documentViewer.annotationManager.createMeasurementAnnotation(annotationData)
    this.annotation = annots[0]

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    this.updateMeasureData()
  }

  async handleMouseMove(event: MouseEvent) {
    if (this.drawingPageIndex < 0) return

    let { x, y } = getAbsoluteCoordinate(this.drawingLayer.pageDiv, event)
    const { width, height, scale } = this.drawingLayer.viewport

    x = Math.max(0, Math.min(width, x))
    y = Math.max(0, Math.min(height, y))

    this.realEndPoint = { x, y }
    if (this.realStartPoint?.x === this.realEndPoint.x) {
      this.realEndPoint.x += 1
    }
    if (this.realStartPoint?.y === this.realEndPoint.y) {
      this.realEndPoint.y += 1
    }

    // 更新测量数据
    const oriPoint = getInitialPoint({ x, y }, scale)
    this.oriEndPoint = { x: oriPoint.x, y: oriPoint.y }
    const start = this.oriStartPoint
    const end = this.oriEndPoint

    const rect = {
      left: Math.min(start.x, end.x),
      top: Math.min(start.y, end.y),
      right: Math.max(start.x, end.x),
      bottom: Math.max(start.y, end.y)
    }
    const newRect = this.transformRect(rect, this.drawingLayer.viewport)
    const vertices = [
      { x: newRect.left, y: newRect.top },
      { x: newRect.right, y: newRect.top },
      { x: newRect.right, y: newRect.bottom },
      { x: newRect.left, y: newRect.bottom }
    ]

    const annotationData = {
      pagePtr: this.annotation.pagePtr,
      annotPtr: this.annotation.annotPtr,
      rulerPtr: this.annotation.rulerPtr,
      pageIndex: this.annotation.pageIndex,
      rect,
      vertices,
    }
    if (this.annotation.contents) {
      annotationData.contents = this.annotation.contents
      annotationData.labelPosition = this.annotation.labelPosition
    }
    this.annotation.rect = rect
    this.annotation.vertices = vertices

    await this.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
    await this.getMeasureData()
    const textList = []
    this.showPerimeter && textList.push(this.annotation.perimeter + this.defaults.scale.translateUnit)
    this.showArea && textList.push(this.annotation.area + this.defaults.scale.translateUnit + '\u00B2')
    this.annotation.contents = textList.join('\n')

    // 绘制元素
    if (!this.shapeEle) {
      this.shapeEle = createSvg('rect', {
        'fill': 'none',
        'stroke': this.defaultStyles.borderColor,
        'stroke-width': 2 * scale,
        'stroke-opacity': 0.8
      })
      this.drawingLayer.svgElement.append(this.shapeEle)
    }

    if (!this.textEle) {
      this.textEle = createSvg('text', {
        'fill': this.defaultStyles.color || this.defaultStyles.borderColor,
        'text-anchor': 'middle',
        'dominant-baseline': 'middle',
        'font-size': this.defaultStyles.fontSize + 'px'
      })
      this.drawingLayer.svgElement.append(this.textEle)
    }

    const midpoint = {
      x: (this.realStartPoint.x + this.realEndPoint.x) / 2 - this.annotation.borderWidth,
      y: (this.realStartPoint.y + this.realEndPoint.y) / 2 - this.annotation.borderWidth
    }

    this.shapeEle.setAttribute('x', Math.min(this.realStartPoint.x, this.realEndPoint.x) - this.annotation.borderWidth,)
    this.shapeEle.setAttribute('y', Math.min(this.realStartPoint.y, this.realEndPoint.y) - this.annotation.borderWidth,)
    this.shapeEle.setAttribute('width', `${Math.abs(this.realEndPoint.x - this.realStartPoint.x) + this.annotation.borderWidth * 2}px`,)
    this.shapeEle.setAttribute('height', `${Math.abs(this.realEndPoint.y - this.realStartPoint.y) + this.annotation.borderWidth * 2}px`,)

    this.textEle.textContent = ''

    if (this.showPerimeter) {
      const tspan1 = createSvg('tspan', {
        x: midpoint.x,
        y: midpoint.y,
        dy: '0em'
      })
      tspan1.textContent = this.annotation.perimeter + this.defaults.scale.translateUnit
      this.textEle.appendChild(tspan1)
    }
    if (this.showArea) {
      const tspan2 = createSvg('tspan', {
        x: midpoint.x,
        y: midpoint.y,
        dy: '1em'
      })
      tspan2.textContent = this.annotation.area + this.defaults.scale.translateUnit + '\u00B2'
      this.textEle.appendChild(tspan2)
    }

    this.updateMeasureData()
  }

  async handleMouseUp(event: MouseEvent) {
    let { x, y } = getAbsoluteCoordinate(this.drawingLayer.pageDiv, event)
    const { width, height, scale } = this.drawingLayer.viewport

    x = Math.max(0, Math.min(width, x))
    y = Math.max(0, Math.min(height, y))

    this.realEndPoint = { x, y }

    this.drawingPageIndex = -1
    const oriPoint = getInitialPoint({ x, y }, scale)
    this.oriEndPoint = { x: oriPoint.x, y: oriPoint.y }

    if (this.realEndPoint && this.realStartPoint?.x !== this.realEndPoint.x && this.realStartPoint?.y !== this.realEndPoint.y) {
      const start = this.oriStartPoint
      const end = this.oriEndPoint

      if (!start) return
      const rect = {
        left: Math.min(start.x, end.x),
        top: Math.min(start.y, end.y),
        right: Math.max(start.x, end.x),
        bottom: Math.max(start.y, end.y)
      }
      const newRect = this.transformRect(rect, this.drawingLayer.viewport)
      const vertices = [
        { x: newRect.left, y: newRect.top },
        { x: newRect.right, y: newRect.top },
        { x: newRect.right, y: newRect.bottom },
        { x: newRect.left, y: newRect.bottom }
      ]

      const annotationData = {
        pagePtr: this.annotation.pagePtr,
        annotPtr: this.annotation.annotPtr,
        rulerPtr: this.annotation.rulerPtr,
        pageIndex: this.annotation.pageIndex,
        rect,
        vertices,
        labelPosition: this.annotation.labelPosition,
        contents: this.annotation.contents
      }
      this.annotation.rect = rect
      this.annotation.vertices = vertices
      this.annotation.showPerimeter = this.showPerimeter
      this.annotation.showArea = this.showArea
      this.annotation.factor = await this.documentViewer.messageHandler.sendWithPromise('GetFactor', this.annotation.rulerPtr)

      await this.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
      await this.documentViewer.annotationManager.drawAnnotationsFromList(this.annotation)
    } else {
      this.documentViewer.deleteAnnotations(this.annotation)
    }

    this.annotation = null
    this.realStartPoint = null
    this.realEndPoint = null
    this.oriStartPoint = null
    this.oriEndPoint = null
    this.shapeEle = null
    this.textEle = null
    this.drawingLayer.svgElement.innerHTML = ''

    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    this.documentViewer.eventBus.dispatch('annotationUpdated', { annotations: this.documentViewer.annotations })
  }

  destroy() {
    this.reset()
    this.eventBus._off('toolChanged', this.handleTool)
  }

  getMouseLocation(event) {
    const scrollElement = this.documentViewer.getScrollViewElement()
    let margin = document.querySelector('.document-container')?.style.marginLeft || 0
    margin = parseFloat(margin)
    const scrollLeft = scrollElement.scrollLeft || 0
    const scrollTop = scrollElement.scrollTop || 0
    const { pageX, pageY } = getClickPoint(event)
    return {
      x: pageX + scrollLeft - margin,
      y: pageY + scrollTop
    }
  }

  async getMeasureData() {
    const { rulerPtr } = this.annotation
    const area = await this.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { rulerPtr, captionType: 'area' })
    const perimeter = await this.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { rulerPtr, captionType: 'length' })

    const formatValue = (value) => (value % 1 === 0 ? value : Number(value.toFixed(this.decimalPlaces)))

    this.annotation.area = formatValue(area)
    this.annotation.perimeter = formatValue(perimeter)
  }

  async updateMeasureData() {
    this.documentViewer.eventBus.dispatch('updateMeasureData', {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      perimeter: this.annotation.perimeter || 0,
      area: this.annotation.area || 0
    })
  }

  async setDefaults(data) {
    if (data.showPerimeter !== undefined || data.showArea !== undefined) {
      this.showPerimeter = data.showPerimeter ?? this.showPerimeter
      this.showArea = data.showArea ?? this.showArea
    }

    Object.entries(data).forEach(([newKey, newValue]) => {
      Object.entries(this.defaults).forEach(([key, value]) => {
        if (newKey === key && value !== newValue && !['showPerimeter', 'showArea'].includes(newKey)) {
          this.defaults[key] = newValue

          if (key === 'precision') {
            this.decimalPlaces = newValue.toString().split('.')[1]?.length || 0
          }
        }
      })
    })

    if (this.annotationStore.selectedElementName) return

    const measureData = {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      perimeter: 0,
      area: 0
    }

    if (this.annotation) {
      await this.getMeasureData()
      const textList = []
      this.showPerimeter && textList.push(this.annotation.perimeter + this.defaults.scale.translateUnit)
      this.showArea && textList.push(this.annotation.area + this.defaults.scale.translateUnit + '\u00B2')
      this.annotation.contents = textList.join('\n')

      if (this.annotation.rulerPtr && this.oriEndPoint) {
        const factor = await this.documentViewer.messageHandler.sendWithPromise('GetFactor', this.annotation.rulerPtr)
        this.annotation.factor = factor
      }
    }

    this.documentViewer.eventBus.dispatch('updateMeasureData', measureData)
  }

  // 转为以左下角为原点
  transformRect(rect, viewport) {
    const { viewBox, rotation } = viewport
    const width = viewBox[2]
    const height = viewBox[3]

    const radians = rotation * Math.PI / 180;

    // 中心点
    const centerX = width / 2;
    const centerY = height / 2;

    // 提取左上角和右下角的坐标
    const startX = rect.left;
    const startY = rect.top;
    const endX = rect.right;
    const endY = rect.bottom;

    // 旋转矩阵
    const cosTheta = Math.cos(radians);
    const sinTheta = Math.sin(radians);

    // 旋转函数
    const rotatePoint = (x, y) => {
      const rotatedX = cosTheta * (x - centerX) - sinTheta * (y - centerY) + centerX;
      const rotatedY = sinTheta * (x - centerX) + cosTheta * (y - centerY) + centerY;
      return { x: rotatedX, y: rotatedY };
    }

    // 旋转左上角和右下角点
    const rotatedStart = rotatePoint(startX, startY);
    const rotatedEnd = rotatePoint(endX, endY);

    // 转换为以左下角为原点
    const left = rotatedStart.x;
    const top = height - rotatedStart.y;
    const right = rotatedEnd.x;
    const bottom = height - rotatedEnd.y;

    return { left, top, right, bottom };
  }
}