// @ts-nocheck
import ArrowHelper from '../annotation/arrow'
import { getAbsoluteCoordinate, createSvg, getClickPoint, getInitialPoint } from '../annotation/utils'

interface Point {
  x: number,
  y: number
}

export default class DistanceMeasurementCreateTool {
  documentViewer: any
  tool: any
  eventBus: any

  measureType: string
  defaults: {
    scale: {
      base: number
      baseUnit: string
      translate: number
      translateUnit: string
    }
    precision: number
    factor: number
  }
  defaultStyles: {
    borderWidth: number
    opacity: number
    color: string
    borderColor: string
    borderStyle: string
    dashes: number[]
    head: string
    tail: string
    fontName: string
    fontSize: number
  }
  leaderLineDefaultOptions: {
    leadLength: number
    leadOffset: number
    leadExtension: number
    labelPosition: string
  }
  drawingPageIndex: number = -1 // 绘制的页面索引，大于等于0代表正在绘制中
  drawingLayer: any = null
  realStartPoint: Point | null = null
  realEndPoint: Point | null = null
  oriStartPoint: Point | null = null
  oriEndPoint: Point | null = null
  startLineEle: HTMLDivElement | null = null
  endLineEle: HTMLDivElement | null = null
  startArrowEle: HTMLDivElement | null = null
  endArrowEle: HTMLDivElement | null = null
  startPerpendicularLine: HTMLDivElement | null = null
  endPerpendicularLine: HTMLDivElement | null = null
  annotation: any | null = null // 正在创建的注释信息
  decimalPlaces: number

  onClick = this.handleClick.bind(this)
  onMousemove = this.handleMouseMove.bind(this)
  handleTool = this.handleTool.bind(this)

  constructor(options: {
    documentViewer: any,
    annotationStore: any
  }) {
    this.documentViewer = options.documentViewer
    this.tool = options.documentViewer.activeTool
    this.eventBus = options.documentViewer.eventBus
    this.annotationStore = options.annotationStore

    this.measureType = 'distance'
    this.defaults = { // 测量属性
      scale: {
        base: 1,
        baseUnit: 'cm',
        translate: 1,
        translateUnit: 'cm',
      },
      precision: 1,
    }
    this.defaultStyles = { // 注释属性
      borderWidth: 2,
      opacity: 1,
      color: '#FF0000',
      borderColor: '#FF0000',
      borderStyle: 'solid',
      dashes: [],
      head: 'OpenArrow',
      tail: 'OpenArrow',
      fontName: 'Helvetica', // Helvetica, Helvetica-Bold, Helvetica-BoldOblique, Courier, Courier-Oblique, Times, Times-Italic
      fontSize: 14,
      fontStyle: 'regular',
      // contents: 'test',
    }
    this.leaderLineDefaultOptions = {
      leadLength: 9,
      leadOffset: -9,
      leadExtension: 6,
      labelPosition: 'inside', // inside / top
    }
    this.decimalPlaces = this.defaults.precision.toString().split('.')[1]?.length || 0 // 保留小数位

    this.eventBus._on('toolChanged', this.handleTool)
  }

  handleTool({
    tool,
    color
  }) {
    const oldtool = this.tool
    this.tool = tool

    if (oldtool === 'line' && this.tool !== oldtool) {
      this.reset()
    }
    if (this.tool === 'line' && this.documentViewer.toolMode === 'measurement') {
      this.init()
    }
  }

  init() {
    document.removeEventListener('mousedown', this.onMousedown)
    document.removeEventListener('mousemove', this.onMousemove)
    // document.removeEventListener('mouseup', this.onMouseup)

    // document.addEventListener('mousedown', this.onMousedown)
    this.documentViewer.viewerContainer.addEventListener('click', this.onClick)

    this.documentViewer.viewerContainer.classList.add('annotation-edit')
  }

  reset() {
    this.realStartPoint = null
    this.realEndPoint = null
    this.oriStartPoint = null
    this.oriEndPoint = null
    this.startLineEle = null
    this.endLineEle = null
    this.startArrowEle = null
    this.endArrowEle = null
    this.startPerpendicularLine = null
    this.endPerpendicularLine = null

    this.documentViewer.viewerContainer.removeEventListener('click', this.onClick)
    document.removeEventListener('mousemove', this.onMousemove)
    // document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('mousedown', this.onMousedown)

    this.documentViewer.viewerContainer.classList.remove('annotation-edit')
  }

  async handleClick(event) {
    const windowCoordinates = this.getMouseLocation(event)
    const page = this.documentViewer.getSelectedPage(windowCoordinates, windowCoordinates)
    const pageIndex = page.first - 1
    if (pageIndex < 0) return
    if (this.annotationStore.selectedElementName && this.drawingPageIndex < 0) return

    const layer = this.documentViewer.pdfViewer._pages[pageIndex].compdfAnnotationLayer
    const realPoint = getAbsoluteCoordinate(layer.pageDiv, event) // 实际坐标
    const oriPoint = this.documentViewer.windowToPage(windowCoordinates, pageIndex + 1) // 原始坐标

    // 开始创建
    if (this.drawingPageIndex < 0) {
      this.drawingPageIndex = pageIndex
      this.drawingLayer = layer
      this.drawingLayer.annotationStore.creating = true
      this.realStartPoint = { x: realPoint.x, y: realPoint.y }
      this.oriStartPoint = { x: oriPoint.x, y: oriPoint.y }

      const start = this.oriStartPoint
      const rect = {
        left: start.x,
        top: start.y,
        right: start.x + 1,
        bottom: start.y + 1
      }
      let annotationData: Record<string, unknown> = {
        measure: 1,
        operate: 'add-annot',
        type: this.tool,
        measureType: this.measureType,
        pageIndex,
        date: new Date(),
        rect,
        ...this.defaults,
        ...this.defaultStyles,
      }

      if (annotationData.type == 'line') {
        const linePoints = [start.x, start.y, start.x + 1, start.y + 1]
        annotationData.linePoints = linePoints
        annotationData = Object.assign(annotationData, this.leaderLineDefaultOptions)
      }
      const annots = await this.documentViewer.annotationManager.createMeasurementAnnotation(annotationData)
      this.annotation = annots[0]

      document.addEventListener('mousemove', this.onMousemove)
      this.updateMeasureData()
      return
    }

    // 结束创建
    this.drawingPageIndex = -1
    this.drawingLayer.annotationStore.creating = false
    this.realEndPoint = { x: realPoint.x, y: realPoint.y }
    this.oriEndPoint = { x: oriPoint.x, y: oriPoint.y }

    if (this.realEndPoint && (this.realStartPoint?.x !== this.realEndPoint.x || this.realStartPoint?.y !== this.realEndPoint.y)) {
      let start = { ...this.oriStartPoint }
      let end = { ...this.oriEndPoint }
      
      if (this.oriStartPoint.y === this.oriEndPoint.y || this.oriStartPoint.x === this.oriEndPoint.x) {
        start.x = Math.min(this.oriEndPoint.x, this.oriStartPoint.x)
        start.y = Math.min(this.oriEndPoint.y, this.oriStartPoint.y)
        end.x = Math.max(this.oriEndPoint.x, this.oriStartPoint.x)
        end.y = Math.max(this.oriEndPoint.y, this.oriStartPoint.y)
      } else if (this.oriStartPoint.y > this.oriEndPoint.y) {
        start = this.oriEndPoint
        end = this.oriStartPoint
      }

      if (!start) return
      const rect = {
        left: Math.min(start.x, end.x),
        top: Math.min(start.y, end.y),
        right: Math.max(start.x, end.x),
        bottom: Math.max(start.y, end.y)
      }
      const linePoints = [start.x, start.y, end.x, end.y]

      const annotationData = {
        pagePtr: this.annotation.pagePtr,
        annotPtr: this.annotation.annotPtr,
        rulerPtr: this.annotation.rulerPtr,
        pageIndex: this.annotation.pageIndex,
        rect,
        linePoints,
        labelPosition: this.annotation.labelPosition,
        contents: this.annotation.contents
      }
      this.annotation.rect = rect
      this.annotation.linePoints = linePoints

      await this.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
      await this.documentViewer.annotationManager.drawAnnotationsFromList(this.annotation)
    } else {
      this.documentViewer.deleteAnnotations(this.annotation)
    }

    this.annotation = null
    this.realStartPoint = null
    this.realEndPoint = null
    this.oriStartPoint = null
    this.oriEndPoint = null
    this.startLineEle = null
    this.endLineEle = null
    this.startArrowEle = null
    this.endArrowEle = null
    this.startPerpendicularLine = null
    this.endPerpendicularLine = null
    layer.svgElement.innerHTML = ''

    document.removeEventListener('mousemove', this.onMousemove)
    this.documentViewer.eventBus.dispatch('annotationUpdated', { annotations: this.documentViewer.annotations })
  }

  async handleMouseMove(event: MouseEvent) {
    if (this.drawingPageIndex < 0) return

    let { x, y } = getAbsoluteCoordinate(this.drawingLayer.pageDiv, event)
    const { width, height, scale } = this.drawingLayer.viewport

    x = Math.max(0, Math.min(width, x))
    y = Math.max(0, Math.min(height, y))

    this.realEndPoint = { x, y }
    if (this.realStartPoint?.x === this.realEndPoint.x) {
      this.realEndPoint.x += 1
    }
    if (this.realStartPoint?.y === this.realEndPoint.y) {
      this.realEndPoint.y += 1
    }

    // 更新测量数据
    const oriPoint = getInitialPoint({ x, y }, scale)
    this.oriEndPoint = { x: oriPoint.x, y: oriPoint.y }
    
    let start = { ...this.oriStartPoint }
    let end = { ...this.oriEndPoint }
    
    if (this.oriStartPoint.y === this.oriEndPoint.y || this.oriStartPoint.x === this.oriEndPoint.x) {
      start.x = Math.min(this.oriEndPoint.x, this.oriStartPoint.x)
      start.y = Math.min(this.oriEndPoint.y, this.oriStartPoint.y)
      end.x = Math.max(this.oriEndPoint.x, this.oriStartPoint.x)
      end.y = Math.max(this.oriEndPoint.y, this.oriStartPoint.y)
    } else if (this.oriStartPoint.y > this.oriEndPoint.y) {
      start = this.oriEndPoint
      end = this.oriStartPoint
    }

    const rect = {
      left: Math.min(start.x, end.x),
      top: Math.min(start.y, end.y),
      right: Math.max(start.x, end.x),
      bottom: Math.max(start.y, end.y)
    }
    const linePoints = [start.x, start.y, end.x, end.y]

    const annotationData = {
      pagePtr: this.annotation.pagePtr,
      annotPtr: this.annotation.annotPtr,
      rulerPtr: this.annotation.rulerPtr,
      pageIndex: this.annotation.pageIndex,
      rect,
      linePoints,
    }
    if (this.annotation.contents) {
      annotationData.contents = this.annotation.contents
      annotationData.labelPosition = this.annotation.labelPosition
    }
    this.annotation.rect = rect
    this.annotation.linePoints = linePoints

    await this.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
    const measureResult = await this.getMeasureData()
    this.annotation.measureResult = measureResult
    this.annotation.contents = measureResult + this.defaults.scale.translateUnit

    // 绘制元素
    if (!this.startLineEle) {
      this.startLineEle = createSvg('line', {
        'stroke': this.defaultStyles.borderColor,
        'stroke-width': 2 * scale,
        'stroke-opacity': 0.8
      })
      this.endLineEle = createSvg('line', {
        'stroke': this.defaultStyles.borderColor,
        'stroke-width': 2 * scale,
        'stroke-opacity': 0.8
      })
      this.drawingLayer.svgElement.append(this.endLineEle)
      this.drawingLayer.svgElement.append(this.startLineEle)

      this.startArrowEle = this.createArrow(this.realStartPoint, this.realEndPoint, true)
      this.endArrowEle = this.createArrow(this.realStartPoint, this.realEndPoint, false)

      this.textEle = createSvg('text', {
        'fill': this.defaultStyles.color || this.defaultStyles.borderColor,
        'text-anchor': 'middle',
        'dominant-baseline': 'middle',
        'font-size': this.defaultStyles.fontSize + 'px'
      })
      this.drawingLayer.svgElement.append(this.textEle)
    }

    this.updateArrow(this.startArrowEle, this.realStartPoint, this.realEndPoint, true)
    this.updateArrow(this.endArrowEle, this.realStartPoint, this.realEndPoint, false)

    const midpoint = {
      x: (this.realStartPoint.x + this.realEndPoint.x) / 2,
      y: (this.realStartPoint.y + this.realEndPoint.y) / 2
    }
    const angleDeg = Math.atan2(this.realEndPoint.y - this.realStartPoint.y, this.realEndPoint.x - this.realStartPoint.x) / Math.PI * 180
    const rotateAngle = Math.abs(angleDeg) > 90 ? angleDeg - 180 : angleDeg
    const angle = Math.abs(angleDeg) > 90 ? 180 - Math.abs(angleDeg) : angleDeg
    const formatAngle = Math.abs(angle).toFixed(this.decimalPlaces) * 1
    this.annotation.angle = formatAngle

    const dx = this.realEndPoint.x - this.realStartPoint.x
    const dy = this.realEndPoint.y - this.realStartPoint.y
    const length = Math.sqrt(dx * dx + dy * dy)
    const unitX = dx / length
    const unitY = dy / length
    const perpX = -unitY
    const perpY = unitX
    const gap = this.annotation.contents.length * this.annotation.fontSize * 0.65
    const textGap = length > gap ? gap : length

    this.startLineEle.setAttribute('x1', `${this.realStartPoint.x}`)
    this.startLineEle.setAttribute('y1', `${this.realStartPoint.y}`)
    this.startLineEle.setAttribute('x2', `${this.realStartPoint.x + unitX * (length / 2 - textGap / 2)}`)
    this.startLineEle.setAttribute('y2', `${this.realStartPoint.y + unitY * (length / 2 - textGap / 2)}`)

    this.endLineEle.setAttribute('x1', `${this.realStartPoint.x + unitX * (length / 2 + textGap / 2)}`)
    this.endLineEle.setAttribute('y1', `${this.realStartPoint.y + unitY * (length / 2 + textGap / 2)}`)
    this.endLineEle.setAttribute('x2', `${this.realEndPoint.x}`)
    this.endLineEle.setAttribute('y2', `${this.realEndPoint.y}`)

    this.textEle.setAttribute('x', midpoint.x)
    this.textEle.setAttribute('y', midpoint.y)
    this.textEle.setAttribute('transform', `rotate(${rotateAngle} ${midpoint.x} ${midpoint.y})`)
    this.textEle.textContent = this.annotation.contents

    // this.updatePerpendicularLines()
    this.updateMeasureData()
  }

  // 创建箭头
  createArrow(startPoint, endPoint, isStart = true) {
    const scale = this.drawingLayer.viewport.scale
    const arrowBuild = new ArrowHelper(startPoint, endPoint)
    const arrowUi = isStart ? arrowBuild.getStartArrow(45, 10 * scale, false) : arrowBuild.getEndArrow(45, 10 * scale, false)

    if (arrowUi) {
      arrowUi.setAttribute('stroke', '#FF0000')
      arrowUi.setAttribute('stroke-width', `${2 * scale}`)
      arrowUi.setAttribute('fill', 'none')
      arrowUi.setAttribute('stroke-opacity', '0.8')
      this.drawingLayer.svgElement.append(arrowUi)
    }

    return arrowUi
  }

  // 更新箭头的位置
  updateArrow(arrow, startPoint, endPoint, isStart) {
    if (!arrow) return

    const arrowBuild = new ArrowHelper(startPoint, endPoint)
    const arrowPoints = isStart ? arrowBuild.getStartArrowPoints(45, 10 * this.drawingLayer.viewport.scale) : arrowBuild.getEndArrowPoints(45, 10 * this.drawingLayer.viewport.scale)

    if (arrowPoints && arrowPoints.length) {
      const points = arrowPoints.map(element => `${element.x},${element.y}`).join(' ')
      arrow.setAttribute('points', points)
    }
  }

  // 更新两端的短线段
  updatePerpendicularLines(lineLength = 12, distanceOffset = 4) {
    const dx = this.realEndPoint.x - this.realStartPoint.x
    const dy = this.realEndPoint.y - this.realStartPoint.y

    const length = Math.sqrt(dx * dx + dy * dy)
    if (length === 0) return

    const ux = dx / length // 单位向量x
    const uy = dy / length // 单位向量y

    const perpUx = -uy // 垂直向量x
    const perpUy = ux  // 垂直向量y

    const scale = this.drawingLayer.viewport.scale
    const halfLineLength = lineLength * scale / 2

    // 计算垂直线段的端点
    const startPerp1 = {
      x: this.realStartPoint.x + perpUx * halfLineLength,
      y: this.realStartPoint.y + perpUy * halfLineLength
    }
    const endPerp1 = {
      x: this.realStartPoint.x - perpUx * halfLineLength,
      y: this.realStartPoint.y - perpUy * halfLineLength
    }

    const startPerp2 = {
      x: this.realEndPoint.x + perpUx * halfLineLength,
      y: this.realEndPoint.y + perpUy * halfLineLength
    }
    const endPerp2 = {
      x: this.realEndPoint.x - perpUx * halfLineLength,
      y: this.realEndPoint.y - perpUy * halfLineLength
    }

    if (this.startPerpendicularLine) {
      this.startPerpendicularLine.setAttribute('x1', startPerp1.x)
      this.startPerpendicularLine.setAttribute('y1', startPerp1.y)
      this.startPerpendicularLine.setAttribute('x2', endPerp1.x)
      this.startPerpendicularLine.setAttribute('y2', endPerp1.y)
    } else {
      this.startPerpendicularLine = createSvg('line', {
        'x1': startPerp1.x,
        'y1': startPerp1.y,
        'x2': endPerp1.x,
        'y2': endPerp1.y,
        'stroke': '#FF0000',
        'stroke-width': 2 * scale,
        'stroke-opacity': 0.8
      })
      this.drawingLayer.svgElement.append(this.startPerpendicularLine)
    }

    if (this.endPerpendicularLine) {
      this.endPerpendicularLine.setAttribute('x1', startPerp2.x)
      this.endPerpendicularLine.setAttribute('y1', startPerp2.y)
      this.endPerpendicularLine.setAttribute('x2', endPerp2.x)
      this.endPerpendicularLine.setAttribute('y2', endPerp2.y)
    } else {
      this.endPerpendicularLine = createSvg('line', {
        'x1': startPerp2.x,
        'y1': startPerp2.y,
        'x2': endPerp2.x,
        'y2': endPerp2.y,
        'stroke': '#FF0000',
        'stroke-width': 2 * scale,
        'stroke-opacity': 0.8
      })
      this.drawingLayer.svgElement.append(this.endPerpendicularLine)
    }
  }

  destroy() {
    this.reset()
    this.eventBus._off('toolChanged', this.handleTool)
  }

  getMouseLocation(event) {
    const scrollElement = this.documentViewer.getScrollViewElement()
    let margin = document.querySelector('.document-container')?.style.marginLeft || 0
    margin = parseFloat(margin)
    const scrollLeft = scrollElement.scrollLeft || 0
    const scrollTop = scrollElement.scrollTop || 0
    const { pageX, pageY } = getClickPoint(event)
    return {
      x: pageX + scrollLeft - margin,
      y: pageY + scrollTop
    }
  }

  getLeaderLineDefaultOptions() {
    return this.leaderLineDefaultOptions
  }

  async getMeasureData() {
    const { annotPtr, rulerPtr } = this.annotation
    const value = await this.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { annotPtr, rulerPtr })

    if (value % 1 === 0) {
      return value
    } else {
      return value.toFixed(this.decimalPlaces)
    }
  }

  async updateMeasureData() {
    let X = 0
    let Y = 0

    if (this.annotation.rulerPtr && this.oriEndPoint) {
      const factor = await this.documentViewer.messageHandler.sendWithPromise('GetFactor', this.annotation.rulerPtr)

      X = (Math.abs(this.oriEndPoint.x - this.oriStartPoint.x) * factor).toFixed(this.decimalPlaces)
      Y = (Math.abs(this.oriEndPoint.y - this.oriStartPoint.y) * factor).toFixed(this.decimalPlaces)

      this.annotation.factor = factor
      this.annotation.X = X
      this.annotation.Y = Y
    }

    this.documentViewer.eventBus.dispatch('updateMeasureData', {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      length: this.annotation.measureResult || 0,
      angle: this.annotation.angle || 0,
      X,
      Y,
    })
  }

  async setDefaults(data) {
    Object.entries(data).forEach(([newKey, newValue]) => {
      Object.entries(this.defaults).forEach(([key, value]) => {
        if (newKey === key && value !== newValue) {
          this.defaults[key] = newValue

          if (key === 'precision') {
            this.decimalPlaces = newValue.toString().split('.')[1]?.length || 0
          }
        }
      })
    })

    if (this.annotationStore.selectedElementName) return

    const measureData = {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      length: 0,
      angle: 0,
      X: 0,
      Y: 0,
    }

    if (this.annotation) {
      const measureResult = await this.getMeasureData()
      this.annotation.measureResult = measureResult
      this.annotation.contents = measureResult + this.defaults.scale.translateUnit
      measureData.measureResult = measureResult
    }

    this.documentViewer.eventBus.dispatch('updateMeasureData', measureData)
  }
}